#ifndef MCL_CORE
#define MCL_CORE

#include <pthread.h>
#include <vector>


namespace mcl {

  /*
   * Virtual class that defines the scheduler interface.
   *
   */
  // class Scheduler {
  //   //forward declaration of nested class handle
  // public:
  //   class Handle;

  //   //class scheduler
  // public:
  //   //GetHandle returns a new scheduler handle.
  //   //Used by ThreadPool during initialisation.
  //   virtual Handle* GetHandle()=0;
  //   //AddTask add a new Task to the scheduler.
  //   //used to add the first task to th scheduler, or to add a seperate top level Task.
  //   virtual void AddTask(Task*)=0;
  //   //IsDone returns true if all scheduled Tasks have been completed.
  //   virtual bool IsDone() = 0;

  //   /*
  //    * Virtual class that define the scheduler handle interface.
  //    * Scheduler handles provide a communication channel between threads and the scheduler.
  //    * Scheduler handles can also be used by the scheduler for storing any per thread data.
  //    */
  //   class Handle {
  //   public:
  //     /*
  //      * Spawn: tell the scheduler to add a number of new Tasks.
  //      * The call to Spawn blocks untill all the spawned Tasks have completed.
  //      * The execution of the Tasks is started in the same order as in the arguments to Spawn,
  //      *   e.g. execution is started on the first argument to spawn before any of the other Tasks in the same Spawn.
  //      * No other guarantees are given by Spawn.
  //      * The Spawned Tasks are assumed to be completely independent
  //      *   and any dependencies between spawned tasks cause  undefined behaviour.
  //      * 
  //      */
  //     //template<class... Tasks>
  //     //virtual void Spawn(Task* t,Tasks... tasks)=0;
  //     //template<class TaskView>
  //     //virtual void Spawn(TaskView tv)=0;
  //     /*
  //      * Get the next task that the thread should run
  //      * TryGetTask is allowed to return NULL, even if there are ready tasks 
  //      * Null should only be returned in rare cases,
  //      *   but it is allowed to simplify the implementation of schedulers
  //      */
  //     virtual Task* TryGetTask()=0;
  //     //IsDone returns whether all tasks in the scheduler have completed
  //     virtual bool IsDone()=0;
  //     inline void SetThread(WorkerThread* thread) {
  // 	thread_ = thread;
  //     }
  //   protected:
  //     //for debugging purposes, should not be used in atual code
  //     WorkerThread* thread_;
  //   };

  // };

  // /*
  //  * Virtual class, defining the task interface.
  //  * This is the class the user defined tasks should inheirent from.
  //  * Only the operator() function should be inplemented,
  //  *   all other function members are implemented by the library.
  //  * The task class is defined by the scheduler choosen and 
  //  */
  // class Task {
  // public:
  //   // Scheduler handle is set by the scheduler handle
  //   //   once execution of the Task begins.
  //   void SetSchedulerHandle(Scheduler::Handle*);
  //   //Run is called by the library to start the execution of the Task.
  //   virtual void operator()()=0;
  // protected:
  //   /* Spawn is implemented by the library
  //    *   and allow the task to Spawn child Tasks.
  //    * Spawn calles the coresponding SchedulerHandle_->Spawn();
  //    */
  //   template<class... Tasks>
  //     virtual void Spawn(Task* t,Tasks... tasks)=0;
  //   template<class TaskView>
  //     virtual void Spawn(TaskView tv)=0;
  // };

  /*
   * WorkerThread provide the actual parallelism.
   * This is a wrapper around the pthread C library that provides a c++ interface.
   * StartThread is used by the pthread start function and calls DoWork.
   * The pthread is not created by the StartThread function.
   * When the scheduler IsDone returns true the WorkerThread exit and joins with the master thread.
   * TODO:
   *      - is a way to stop a thread needed?
   *      - is a way to abort a task needed?
   */
  template<class Scheduler>
  class WorkerThread {
  public:
    WorkerThread(size_t id, typename Scheduler::Handle* handle) :
      id_(id), schedulerHandle_(handle) {}
    inline void Start(){
      pthread_attr_t attr;
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
      pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
      pthread_create(&pthread_, &attr, &WorkerThread::StartThread, this);
    }
    inline void Join(){
      void* status;
      pthread_join(pthread_,&status);
    }
    inline size_t GetID() {
      return id_;
    }
  private:
    static inline void* StartThread(void* workerThread){
      static_cast<WorkerThread<Scheduler>* >(workerThread)->DoWork();
      return 0;
    }
    inline void DoWork(){
      while(!schedulerHandle_->IsDone()) {
	typename Scheduler::Task* task = schedulerHandle_->TryGetTask();
	if (task == NULL) {
	  //we did not get a task, try again
	  continue;
	}
	task->Execute();
      }
    }
    pthread_t pthread_;
    size_t id_;
    typename Scheduler::Handle* schedulerHandle_;
    //Disable assignment
    WorkerThread(const WorkerThread&);
    void operator=(const WorkerThread&);
  };

  /*
   * ThreadPool creates the WorkerThreads and
   *    waits until all WorkerThreads have joined before exiting.
   */
  template<class Scheduler>
  class ThreadPool {
  public:
    inline ThreadPool(Scheduler& scheduler, size_t numThreads) : numThreads_(numThreads), scheduler_(scheduler) {
      for (size_t i =0; i<numThreads_; ++i) {
	threadArray_.push_back(new WorkerThread<Scheduler>((size_t)i, scheduler_.GetHandle()));
      }
    }
    inline void AddTask(typename Scheduler::Task* task){
      scheduler_.AddTask(task);
    }
    inline void Start() {
      typename std::vector<WorkerThread<Scheduler>*>::iterator it = threadArray_.begin();
      while(it != threadArray_.end()) {
	(*it)->Start();
	++it;
      }
      it = threadArray_.begin();
      while(it != threadArray_.end()) {
	(*it)->Join();
	++it;
      }
    }
  private:
    size_t numThreads_;
    typename std::vector<WorkerThread<Scheduler>*> threadArray_;
    Scheduler& scheduler_;
    //diable copy construction and assignment
    ThreadPool(const ThreadPool&);
    void operator=(const ThreadPool&);
  };

}

#endif //MCL_CORE
