#ifndef MCL_LOCKING_PARALLEL_DEPTH_FIRST_SCHEDULER
#define MCL_LOCKING_PARALLEL_DEPTH_FIRST_SCHEDULER

#include <list>
#include <functional>
#include <stdexcept>

#include <tbb/atomic.h>

#include "MCL/mutex.h"
#include "MCL/scheduling/pdf/lockingTaskQueue.h"
#include "MCL/data_structures/transformView.h"

namespace mcl{

  class LockingPDFScheduler {
  public:
    class Task;
    class Handle;
    
    inline LockingPDFScheduler() : queue_(), handles_() {}
    inline Handle* GetHandle(){
      Handle* handle = new Handle(*this);
      handles_.insert(handles_.begin(),handle);
      return handle;
    }
    void AddTask(Task* task){
      queue_.AddTask(task);
    }
    bool IsDone(){
      return queue_.IsEmpty();
    }
    //Scheduler private members
  private:
    typedef LockingTaskQueue<Task> Queue;
    inline Queue::Handle GetQueueHandle(Task* t){
      return queue_.GetHandle(t);
    }
    Queue queue_;
    std::list<Handle*> handles_;
    //disable copy construction and assignement
    LockingPDFScheduler(const LockingPDFScheduler&);
    void operator=(const LockingPDFScheduler&);

  public:
    class Handle {
      //Handle
    public:
      inline Handle(LockingPDFScheduler& s)
	: scheduler_(s), currentQueueHandle_(NULL,&scheduler_.queue_) {
	//std::cout<<"new pdf::Handle"<<std::endl;
      }
      template<class... Tasks>
      inline void SpawnTasks(Tasks... tasks) {
	//std::cout<<"pdf::Handle::Spawn"<<std::endl;
	//the queue handle changes at insertion, we need to hold on to the parent task.
	Task* parent = currentQueueHandle_.GetTask();
	currentQueueHandle_.InsertTasks(tasks...);
	BlockingRun(parent);
      }
      template<class TaskView>
      void SpawnView(TaskView& taskView){
	Task* parent = currentQueueHandle_.GetTask();
	currentQueueHandle_.InsertView(taskView);
	BlockingRun(parent);
	//std::cout<<"spawn View after blockingrun "<<&currentQueueHandle_<<"="<<currentQueueHandle_<<std::endl;
      }
      inline Task* TryGetTask(Task* parent=0){
	//std::cout<<"pdf::Handle::TryGetTask"<<std::endl;
	currentQueueHandle_  = scheduler_.GetQueueHandle(parent);
	Task* task = currentQueueHandle_.GetTask();
	if(task != NULL)
	  task->SetSchedulerHandle(this);
	//std::cout<<"TryGetTask returning "<<task<<std::endl;
	return task;
      }
      inline bool IsDone(){
	//std::cout<<"pdf::Handle::IsDone"<<std::endl;
	return scheduler_.queue_.IsEmpty();
      }
      inline void TaskDone(){
	currentQueueHandle_.Remove();
      }
    private:
      inline void BlockingRun(Task* parent){
	//std::cout<<"pdf::Handle::BlockingRun("<<parent<<")"<<std::endl;
	Queue::Handle parentHandle = currentQueueHandle_;
	while(!parent->IsReady()){
	  //std::cout<<"waiting on: "<<parent<<std::endl;
	  Task* task = TryGetTask(parentHandle.GetTask());
	  if(task != NULL){
	    task->Execute();
	    //std::cout<<"blockingrun("<<parent<<") restting handel "<<currentQueueHandle_<<" to:"<<parentHandle<<std::endl;
	  } else {
	    //sleep(1);
	  }
	  currentQueueHandle_ = parentHandle;
	}
	//std::cout<<"end of pdf::Handle::BlockingRun("<<parent<<")"<<std::endl;
      }
      LockingPDFScheduler& scheduler_;
      Queue::Handle currentQueueHandle_;
      //disable copy construction and assignement
      Handle(const Handle&);
      void operator=(const Handle&);
    };

    class Task {
    public:
      inline Task() 
	: numSyncedChildren_(), numSpawnedChildren_(0), parent_(0), schedulerHandle_(0) {
	numSyncedChildren_ = 0;
      }
      virtual ~Task() {}
      template<class... Tasks>
      inline void Spawn(Tasks... tasks){
	//std::cout<<this<<" spawning "<<sizeof...(tasks)<<" tasks"<<std::endl;
	numSyncedChildren_ =0;
	numSpawnedChildren_ = sizeof...(tasks);
	SetParentToThis(tasks...);
	schedulerHandle_->SpawnTasks(tasks...);
	numSpawnedChildren_ = 0;
	numSyncedChildren_ = 0;
      }
      inline void Spawn(Task* task){
	numSpawnedChildren_ = 1;
	numSyncedChildren_ = 0;
	task->SetParent(this);	
	schedulerHandle_->SpawnTasks(task);
	numSpawnedChildren_ = 0;
	numSyncedChildren_ = 0;
      }
      template<class TaskView>
      void Spawn(TaskView& taskView){
	//std::cout<<this<<" spawning view with "<<taskView.size()<<" tasks"<<std::endl;
	numSyncedChildren_ =0;
	numSpawnedChildren_ = taskView.size();
	std::function<Task*(Task*)> setParentTransform = 
	  std::bind(&Task::SetParentWrapper,this,std::placeholders::_1);
	TransformView<TaskView,Task*> viewWrapper(&taskView,setParentTransform);
	// Task* t = setParentTransform(*taskView.begin());	
	// typename TransformView<TaskView,Task*>::iterator it = viewWrapper.begin();
	// std::cout<<*it<<std::endl;
	// typename TransformView<TaskView,Task*>::iterator end = viewWrapper.end();
	// while(it != end){
	//   std::cout<<*it<<std::endl;
	//   ++it;
	// }

	schedulerHandle_->SpawnView(viewWrapper);
	numSpawnedChildren_ = 0;
	numSyncedChildren_ = 0;
      }
      inline bool IsReady(){
	unsigned int t = numSyncedChildren_;
	//std::cout<<"Task is Ready spawned: "<<numSpawnedChildren_<<" synced: "<<t<<std::endl;
	return numSpawnedChildren_ == t;
      }
      virtual void operator()() = 0;
      virtual void PostRun() {}
      inline void Execute(){
	//hold on to parent as we can risk that the task is deleted in postrun
	Task* parent = parent_;
	Handle* handle = schedulerHandle_;
	//std::cout<<"Task "<<this<<" Executing"<<std::endl;
	operator()();
	//std::cout<<"Task: "<<this<<" calling PostRun"<<std::endl;
	PostRun();
	if (parent != NULL){
	  //std::cout<<"Task: "<<this<<" notifying parent "<<parent_<<std::endl;
	  parent->IncNumSyncedChildren();
	}
	//std::cout<<"Task: "<<this<<" calling clean"<<std::endl;
	handle->TaskDone();
      }
      inline void SetSchedulerHandle(Handle* h) {
	schedulerHandle_ = h;
      }
      inline Task* GetParent(){
	return parent_;
      }
      inline void SetParent(Task* p){
	parent_ = p;
      }
      inline Task* SetParentWrapper(Task* p){
	p->parent_ = this;
	return p;
      }
      inline void IncNumSyncedChildren(){
	++numSyncedChildren_;
      }
    protected:
      template<class... Tasks>
      inline void SetParentToThis() {} //recursion basis
      template<class... Tasks>
      inline void SetParentToThis(Task& head,Tasks... tail){
	head.SetParent(this);
	SetParentToThis(tail...);
      }
      template<class... Tasks>
      inline void SetParentToThis(Task* head,Tasks... tail){
	if(head != NULL)
	  head->SetParent(this);
	SetParentToThis(tail...);
      }
      tbb::atomic<unsigned int> numSyncedChildren_;
      unsigned int numSpawnedChildren_;
      Task* parent_;
      Handle* schedulerHandle_;
    };
    
  };

}//end namespace

#endif //header guard


